/*  ddtp.c - DD-ternary logic permutation utility

    Part of NedoPC SDK (software development kit for simple devices)

    Copyright (C) 2010, Alexander A. Shabarshin <ashabarshin@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 History of DDTp:
 
   v0.3 (25 July 2010) - first version for DDT v0.3
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "my_text.h"
#include "ddt.h"

//#define DEBUG
//#define DEBUG_

#define STRLEN 1000

void printa(int *ia, int na)
{
  int i;
  for(i=0;i<na;i++)
  {
    if(i!=0) printf("|");
    printf("%i",ia[i]);
  }
}

int correcta(int *ia, int na)
{
  int i,j,k=1;
  for(i=0;i<na;i++)
  {
    for(j=0;j<na;j++)
    {
      if(i!=j && ia[i]==ia[j])
      {
        k = 0;
        break;
      }
    }
    if(!k) break;
  }
#ifdef DEBUG_
  printa(ia,na);printf(" -> %i\n",k);
#endif
  return k;
}

/* Main function */

int main(int argc, char **argv)
{
  FILE *f;
  Line *l;
  Text *t = NULL;
  int i,j,k,m=0,err=0,first=1,*ins=NULL;
  int fun,e12,e21,dgs,orid=-1,maxd=-1;
  char str[STRLEN],*po;
  printf("\nDDTp v" DDT_VERSION " (c) 2010, Alexander A. Shabarshin <ashabarshin@gmail.com>\n\n");
  if(argc < 2)
  {
    printf("Usage:\n\tddtp file.ddt\n\n");
    return -1;
  }
  if(strstr(argv[1],".ddt")==NULL)
  {
    printf("ERROR: %s is not ddt file!\n",argv[1]);
    return -2;
  }
  f = fopen(argv[1],"rt");
  if(f==NULL)
  {
    printf("ERROR: %s can not be opened!\n",argv[1]);
    return -3;
  }
  while(1)
  {
    fgets(str,STRLEN,f);
    if(feof(f)) break;
    po = strrchr(str,'\n');
    if(po!=NULL)
    {
      po = strchr(str,'=');
      if(po!=NULL)
      {
        if(m==0) m=po-str;
        else if(m!=po-str) err=-4;
      }
      else err=-4;
    }
    if(po==NULL || err)
    {
      po = strrchr(str,'\n');
      if(po!=NULL) *po=0;
      po = strrchr(str,'\r');
      if(po!=NULL) *po=0;
      printf("ERROR: line %s is not correct (%i)!\n",str,m);
      err = -4;
      break;
    }
  }
  fclose(f);
  if(!err)
  {
    printf("%i inputs detected\n",m);
    ins = (int*)malloc(sizeof(int)*m);
    t = TextNew();
    if(ins==NULL || t==NULL)
    {
      printf("ERROR: out of memory!\n");
      err = -5;
    }
    else
    {
      TextLoad(t,argv[1]);
      for(i=0;i<m;i++) ins[i]=m-i-1;
    }
  }
  printf("\n");
  while(!err)
  {
#ifdef DEBUG
    printa(ins,m);printf("\n");
#endif
    f = fopen("0.ddt","wt");
    if(f==NULL)
    {
      printf("ERROR: output file can not be created!\n");
      err = -6;
      break;
    }
    for(l=t->first;l!=NULL;l=l->next)
    {
      for(i=0;i<m;i++)
      {
        str[i] = l->str[m-ins[i]-1];
      }
      str[i] = 0;
      fprintf(f,"%s%s\n",str,&l->str[m]);
    }
    fclose(f);
    unlink("ddt_0.c");
    system("ddtc 0.ddt >0.out");
    f = fopen("0.out","rt");
    if(f==NULL)
    {
      printf("ERROR: output file can not be opened!\n");
      err = -7;
      break;
    }
    k = 0;
    fun = 0;
    while(1)
    {
      fgets(str,STRLEN,f);
      if(feof(f)) break;
      if(!strncmp(str,"Generate",8))
      {
        po = strstr(str,"with");
        if(po!=NULL) k = atoi(&po[5]);
      }
      if(!strncmp(str,"OK",2))
      {
        fun = k;
      }
      if(!strncmp(str,"ERR",3))
      {
        printf("ERROR: found error!\n");
        err = -8;
      }
    }
    fclose(f);
    if(err) break;
    e12 = e21 = 0;
    f = fopen("ddt_0.c","rt");
    if(f==NULL)
    {
      printf("ERROR: generated file can not be opened!\n");
      err = -9;
      break;
    }
    while(1)
    {
     fgets(str,STRLEN,f);
     if(feof(f)) break;
     if(strstr(str,"ddt_mux")){e12++;e21++;}
     else if(strstr(str,"ddt_e12")){e12++;}
     else if(strstr(str,"ddt_e21")){e21++;}
     else if(strstr(str,"ddt_blp")){e12++;}
     else if(strstr(str,"ddt_bln")){e21++;}
     else if(strstr(str,"ddt_rup")){e12++;e21++;}
     else if(strstr(str,"ddt_rdn")){e12++;e21++;}
     else if(strstr(str,"ddt_not")){e12++;e21++;}
    }     
    fclose(f);
    i = e12;
    if(i&1) i++;
    j = e21;
    if(j&1) j++;
    dgs = (i+j)>>1;
#ifdef DEBUG
    printf("fun=%i e12=%i e21=%i DG403=%i\n",fun,e12,e21,dgs);
#endif
    if(first)
    {
      maxd = dgs;
      orid = dgs;
      first = 0;
    }
    else
    {
      if(dgs < maxd)
      {
        maxd = dgs;
        printf("NEW ");
        printa(ins,m);
        printf(" e12=%i e21=%i DG403=%i\n",e12,e21,dgs);
        strcpy(str,argv[1]);
        po = strchr(str,'.');
        if(po!=NULL) *po=0;
        strcat(str,"_.ddt");
        rename("0.ddt",str);
      }
    }
    do {
     for(i=m-1;i>=0;i--)
     {
      if(--ins[i]<0 && i>0)
        ins[i] = m-1;
      else break;
     }
     if(ins[0]<0) break;
    }while(!correcta(ins,m));
    if(ins[0]<0) break;
  }
  if(t!=NULL) TextDel(t);
  if(ins!=NULL) free(ins);
  if(orid!=maxd)
  {
    printf("\nOptimal solution has %i chips instead of %i\n\n",maxd,orid);
  }
  else
  {
    printf("\nExisting solution with %i chips is optimal\n\n",orid);
  }
  return err;
}
